/*
 * Copyright 2019 Confluent Inc.
 */

package io.confluent.connect.secretregistry;

import java.io.IOException;
import java.lang.reflect.Method;
import java.security.Principal;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import javax.ws.rs.container.ContainerRequestContext;
import javax.ws.rs.container.ContainerRequestFilter;
import javax.ws.rs.container.ContainerResponseContext;
import javax.ws.rs.container.ContainerResponseFilter;
import javax.ws.rs.container.ResourceInfo;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.SecurityContext;

import org.apache.commons.lang3.tuple.Pair;
import org.apache.kafka.common.security.auth.KafkaPrincipal;
import org.apache.kafka.connect.errors.ConnectException;
import org.apache.kafka.connect.health.ConnectClusterState;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import io.confluent.connect.secretregistry.rbac.SecretRegistryActions;
import io.confluent.connect.secretregistry.rbac.SecretRegistryOperations;
import io.confluent.kafka.secretregistry.rest.resources.PathResource;
import io.confluent.security.auth.client.RestAuthorizer;
import io.confluent.security.authorizer.Action;
import io.confluent.security.authorizer.AuthorizeResult;
import io.confluent.security.authorizer.Scope;

public class ConnectSecretRegistryFilter
    implements ContainerRequestFilter, ContainerResponseFilter {

  private static final Logger log = LoggerFactory.getLogger(ConnectSecretRegistryFilter.class);

  static final Method LIST_PATHS_METHOD;

  static {
    try {
      LIST_PATHS_METHOD = PathResource.class.getMethod("list", SecurityContext.class);
    } catch (NoSuchMethodException e) {
      throw new ConnectException("Failed to initialize Secret Registry RBAC filter", e);
    }
  }

  @Context
  @SuppressWarnings("unused")
  private ResourceInfo resourceInfo;

  private final Scope scope;
  private final SecretRegistryActions actions;
  private final RestAuthorizer restAuthorizer;

  public ConnectSecretRegistryFilter(
      Scope scope,
      RestAuthorizer restAuthorizer,
      ConnectClusterState clusterState
  ) {
    this(scope, SecretRegistryActions.build(scope, clusterState),
        restAuthorizer, null);
  }

  ConnectSecretRegistryFilter(
      Scope scope,
      SecretRegistryActions actions,
      RestAuthorizer restAuthorizer,
      ResourceInfo resourceInfo
  ) {
    this.scope = scope;
    this.restAuthorizer = restAuthorizer;
    this.actions = actions;
    this.resourceInfo = resourceInfo;
  }

  @Override
  public void filter(ContainerRequestContext requestContext) throws IOException {
    if (restAuthorizer == null) {
      log.warn("Rest authorizer not active");
      return;
    }
    Method resourceMethod = resourceInfo.getResourceMethod();
    if (LIST_PATHS_METHOD.equals(resourceMethod)) {
      // Response will be filtered at a later point
      return;
    }
    List<Action> actions = this.actions.actions(resourceMethod, requestContext);
    if (!actions.isEmpty()) {
      Principal requestPrincipal = requestContext.getSecurityContext().getUserPrincipal();
      if (requestPrincipal == null) {
        requestContext.abortWith(Response
            .status(Response.Status.UNAUTHORIZED)
            .entity("User not authenticated")
            .build()
        );
        return;
      }
      KafkaPrincipal userPrincipal = kafkaPrincipalFor(requestPrincipal);
      log.debug("Authorizing request for principal {}. Actions: {}", userPrincipal, actions);
      List<AuthorizeResult> authorizations =
          restAuthorizer.authorize(userPrincipal, null, actions);
      if (!authorizations.stream().allMatch(AuthorizeResult.ALLOWED::equals)) {
        requestContext.abortWith(Response
            .status(Response.Status.FORBIDDEN)
            .entity("Unauthorized operation")
            .build()
        );
      }
    }
  }

  @Override
  public void filter(
      ContainerRequestContext requestContext,
      ContainerResponseContext responseContext
  ) {
    if (restAuthorizer == null) {
      log.warn("Rest authorizer not active");
      return;
    }
    Method resourceMethod = resourceInfo.getResourceMethod();
    if (responseContext.getStatus() == Response.Status.OK.getStatusCode()
        && LIST_PATHS_METHOD.equals(resourceMethod)
    ) {
      Collection<String> allPaths = responseEntity(responseContext);

      if (allPaths.isEmpty()) {
        return;
      }

      List<Action> actions = allPaths.stream()
          .flatMap(path -> SecretRegistryOperations.ALL.stream()
              .map(op -> new Action(
                  scope,
                  SecretRegistryActions.SECRET_RESOURCE,
                  path,
                  op
              )
          )).collect(Collectors.toList());

      List<AuthorizeResult> authorizations = restAuthorizer.authorize(
          kafkaPrincipalFor(requestContext.getSecurityContext().getUserPrincipal()),
          null,
          actions
      );

      Collection<String> visiblePaths = zip(authorizations, actions)
          .filter(pair -> AuthorizeResult.ALLOWED.equals(pair.getLeft()))
          .map(Pair::getRight)
          .map(Action::resourceName)
          .collect(Collectors.toSet());

      responseContext.setEntity(visiblePaths);
    }
  }

  @SuppressWarnings("unchecked")
  private static <T> T responseEntity(ContainerResponseContext responseContext) {
    return (T) responseContext.getEntity();
  }

  /**
   * Generate a {@link KafkaPrincipal} for the given user principal
   * @param userPrincipal the user principal; may not be null
   * @return a {@link KafkaPrincipal} of type {@link KafkaPrincipal#USER_TYPE} and the same name as
   *         the principal
   */
  public static KafkaPrincipal kafkaPrincipalFor(Principal userPrincipal) {
    Objects.requireNonNull(userPrincipal, "Principal may not be null");
    return new KafkaPrincipal(KafkaPrincipal.USER_TYPE, userPrincipal.getName());
  }

  /**
   * Zips two lists together, discarding elements if one is longer than the other.
   * @param left the first list to zip
   * @param right the second list to zip
   * @param <L> the type of the first list
   * @param <R> the type of the second list
   * @return a {@link Stream} of {@link Pair} objects, each representing a pair of elements
   *         which are in the same index of the two lists. The {@link Pair#getLeft()} left} of
   *         each pair contains an element from the first list, and the
   *         {@link Pair#getRight()} right} contains an element from the second list.
   */
  public static <L, R> Stream<Pair<L, R>> zip(List<L> left, List<R> right) {
    return IntStream.range(0, Math.min(left.size(), right.size()))
        .mapToObj(i -> Pair.of(left.get(i), right.get(i)));
  }
}
